package dev.fitko.fitconnect.client.cmd;

import com.beust.jcommander.JCommander;
import com.beust.jcommander.ParameterException;
import dev.fitko.fitconnect.api.domain.model.metadata.attachment.AttachmentWithData;
import dev.fitko.fitconnect.client.SenderClient;
import dev.fitko.fitconnect.client.SubscriberClient;
import dev.fitko.fitconnect.client.cmd.commands.ListAllSubmissionsCommand;
import dev.fitko.fitconnect.client.cmd.commands.ListOneSubmissionCommand;
import dev.fitko.fitconnect.client.cmd.commands.SendSubmissionCommand;
import dev.fitko.fitconnect.client.cmd.util.AttachmentDataType;
import dev.fitko.fitconnect.client.model.ReceivedSubmission;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.List;
import java.util.stream.Collectors;

public class CommandLineClient {

    private static final Logger LOGGER = LoggerFactory.getLogger(CommandLineClient.class);

    private final SendSubmissionCommand sendSubmissionCommand;
    private final ListAllSubmissionsCommand listAllSubmissionsCommand;
    private final ListOneSubmissionCommand listOneSubmissionCommand;

    private final JCommander jc;

    private final SenderClient.Start senderClient;
    private final SubscriberClient.RequestSubmission subscriberClient;

    public CommandLineClient(final SenderClient.Start senderClient,
                             final SubscriberClient.RequestSubmission subscriberClient) {

        this.sendSubmissionCommand = new SendSubmissionCommand();
        this.listAllSubmissionsCommand = new ListAllSubmissionsCommand();
        this.listOneSubmissionCommand = new ListOneSubmissionCommand();

        this.senderClient = senderClient;
        this.subscriberClient = subscriberClient;
        jc = getJCommander();
    }

    public void run(final String... args) throws IOException {
        final var parsedCommand = parseCommandFromArguments(args);
        if (parsedCommand == null) {
            jc.usage();
            return;
        }
        executeCommand(parsedCommand);
    }

    private JCommander getJCommander() {
        return JCommander.newBuilder()
                .addCommand(sendSubmissionCommand)
                .addCommand(listAllSubmissionsCommand)
                .addCommand(listOneSubmissionCommand)
                .build();
    }

    private void executeCommand(final String parsedCommand) throws IOException {
        switch (parsedCommand) {
            case SendSubmissionCommand.SEND_CMD_NAME:
                sendSubmission(sendSubmissionCommand);
                break;
            case ListAllSubmissionsCommand.LIST_CMD_NAME:
                listSubmissions(listAllSubmissionsCommand);
                break;
            case ListOneSubmissionCommand.GET_CMD_NAME:
                listOneSubmission(listOneSubmissionCommand);
                break;
            default:
                break;
        }
    }

    private String parseCommandFromArguments(final String[] args) {
        try {
            jc.parse(args);
            return jc.getParsedCommand();
        } catch (final ParameterException e) {
            LOGGER.error(e.getMessage(), "\n");
            return null;
        }
    }

    private void listOneSubmission(final ListOneSubmissionCommand listOneSubmissionCommand) throws IOException {
        LOGGER.info("Listing submission for id {}", listOneSubmissionCommand.submissionId);
        final var submission = subscriberClient.requestSubmission(listOneSubmissionCommand.submissionId);
        if (submission.isEmpty()) {
            LOGGER.info("No submission found for submission id {}", listOneSubmissionCommand.submissionId);
        } else {
            writeData(submission.get(), getTargetFolderPath(listOneSubmissionCommand));
        }
    }

    private void listSubmissions(final ListAllSubmissionsCommand listAllSubmissionsCommand) {
        LOGGER.info("Listing available submissions for destination {}", listAllSubmissionsCommand.destinationId);
        subscriberClient.getAvailableSubmissions(listAllSubmissionsCommand.destinationId)
                .forEach(submission -> System.out.println(submission.getSubmissionId()));
    }

    private void sendSubmission(final SendSubmissionCommand sendSubmissionCommand) throws IOException {
        LOGGER.info("Sending new submission to destination {}", sendSubmissionCommand.destinationId);
        final List<File> attachments = sendSubmissionCommand.attachments.stream().map(File::new).collect(Collectors.toList());
        if (sendSubmissionCommand.dataType == AttachmentDataType.JSON) {
            sendWithJsonData(sendSubmissionCommand, attachments);
        } else if (sendSubmissionCommand.dataType == AttachmentDataType.XML) {
            sendWithXmlData(sendSubmissionCommand, attachments);
        }
    }

    private void sendWithJsonData(final SendSubmissionCommand sendSubmissionCommand, final List<File> attachments) throws IOException {
        senderClient.newSubmission()
                .withAttachments(attachments)
                .withJsonData(getDataAsString(sendSubmissionCommand.data))
                .withDestination(sendSubmissionCommand.destinationId)
                .withServiceType(sendSubmissionCommand.serviceName, sendSubmissionCommand.leikaKey)
                .submit();
    }

    private void sendWithXmlData(final SendSubmissionCommand sendSubmissionCommand, final List<File> attachments) throws IOException {
        senderClient.newSubmission()
                .withAttachments(attachments)
                .withXmlData(getDataAsString(sendSubmissionCommand.data))
                .withDestination(sendSubmissionCommand.destinationId)
                .withServiceType(sendSubmissionCommand.serviceName, sendSubmissionCommand.leikaKey)
                .submit();
    }

    private String getTargetFolderPath(final ListOneSubmissionCommand listOneSubmissionCommand) {
        if (listOneSubmissionCommand.targetFolder != null) {
            return listOneSubmissionCommand.targetFolder + "/" + listOneSubmissionCommand.submissionId;
        } else {
            return listOneSubmissionCommand.submissionId.toString();
        }
    }

    private String getDataAsString(final String dataPath) throws IOException {
        LOGGER.info("Reading data from {}", dataPath);
        return Files.readString(Path.of(dataPath));
    }

    private void writeData(final ReceivedSubmission submissionData, final String dataDirPath) throws IOException {
        LOGGER.info("Creating data directory for submission in {}", dataDirPath);
        Files.createDirectories(Path.of(dataDirPath));

        final var fileEnding = AttachmentDataType.getFileTypeFromMimeType(submissionData.getMimeType());
        final var filePath = Path.of(dataDirPath + "/data." + fileEnding);
        LOGGER.info("Writing data.{}", fileEnding);
        Files.write(filePath, submissionData.getData().getBytes(StandardCharsets.UTF_8));

        for (final AttachmentWithData attachment : submissionData.getAttachments()) {
            final String filename = attachment.getAttachmentMetadata().getFilename();
            LOGGER.info("Writing attachment {}", filename);
            Files.write(Path.of(dataDirPath, filename), attachment.getDecryptedData());
        }
    }
}
